@@ -36,7 +36,8 @@ class Grilo(GObject.GObject):
 
     __gsignals__ = {
         'ready': (GObject.SIGNAL_RUN_FIRST, None, ()),
-        'changes-pending': (GObject.SIGNAL_RUN_FIRST, None, ())
+        'changes-pending': (GObject.SIGNAL_RUN_FIRST, None, ()),
+        'new-source-added': (GObject.SIGNAL_RUN_FIRST, None, (Grl.Source, ))
     }
 
     METADATA_KEYS = [
@@ -44,6 +45,7 @@ class Grilo(GObject.GObject):
         Grl.METADATA_KEY_ARTIST, Grl.METADATA_KEY_ALBUM,
         Grl.METADATA_KEY_DURATION,
         Grl.METADATA_KEY_CREATION_DATE,
+        Grl.METADATA_KEY_URL,
         Grl.METADATA_KEY_THUMBNAIL]
 
     METADATA_THUMBNAIL_KEYS = [
@@ -70,11 +72,15 @@ class Grilo(GObject.GObject):
                                     Grl.ResolutionFlags.IDLE_RELAY)
 
         self.sources = {}
+        self.blacklist = ['grl-filesystem', 'grl-bookmarks', 'grl-metadata-store', 'grl-podcasts']
         self.tracker = None
         self.changed_media_ids = []
         self.pending_event_id = 0
-
+        self.changes_pending = {'Albums': False, 'Artists': False, 'Songs': False}
         self.registry = Grl.Registry.get_default()
+
+    @log
+    def _find_sources(self):
         self.registry.connect('source_added', self._on_source_added)
         self.registry.connect('source_removed', self._on_source_removed)
 
@@ -95,7 +101,7 @@ class Grilo(GObject.GObject):
                     query = "select DISTINCT rdf:type nie:mimeType(?urn) as mime-type" +\
                             " { ?urn rdf:type nie:InformationElement . FILTER (tracker:id(?urn) = %s) }" % media_id
                     mimeType = grilo.tracker.query_sync(query, [Grl.METADATA_KEY_MIME], grilo.options)[0].get_mime()
-                    if mimeType.startswith("audio"):
+                    if mimeType and mimeType.startswith("audio"):
                         self.changed_media_ids.append(media_id)
                 if changeType == Grl.SourceChangeType.REMOVED:
                     # There is no way to check that removed item is a media
@@ -117,6 +123,9 @@ class Grilo(GObject.GObject):
     def emit_change_signal(self):
         self.changed_media_ids = []
         self.pending_event_id = 0
+        self.changes_pending['Albums'] = True
+        self.changes_pending['Artists'] = True
+        self.changes_pending['Songs'] = True
         self.emit('changes-pending')
         return False
 
@@ -125,19 +134,31 @@ class Grilo(GObject.GObject):
         id = mediaSource.get_id()
         logger.debug("new grilo source %s was added" % id)
         try:
-            if id == 'grl-tracker-source':
-                ops = mediaSource.supported_operations()
-                if ops & Grl.SupportedOps.SEARCH:
-                    print('Detected new source available: \'%s\' and it supports search' %
-                          mediaSource.get_name())
+            ops = mediaSource.supported_operations()
 
+            if id == 'grl-tracker-source':
+                if ops & Grl.SupportedOps.SEARCH:
+                    logger.debug("found searchable tracker source")
                     self.sources[id] = mediaSource
                     self.tracker = mediaSource
+                    self.search_source = mediaSource
 
                     if self.tracker is not None:
                         self.emit('ready')
                         self.tracker.notify_change_start()
                         self.tracker.connect('content-changed', self._on_content_changed)
+
+            elif (id.startswith('grl-upnp')):
+                logger.debug("found upnp source %s" % id)
+                self.sources[id] = mediaSource
+                self.emit('new-source-added', mediaSource)
+
+            elif (id not in self.blacklist) and (ops & Grl.SupportedOps.SEARCH) and \
+                 (mediaSource.get_supported_media() & Grl.MediaType.AUDIO):
+                logger.debug("source %s is searchable" % id)
+                self.sources[id] = mediaSource
+                self.emit('new-source-added', mediaSource)
+
         except Exception as e:
             logger.debug("Source %s: exception %s" % (id, e))
 
@@ -147,42 +168,60 @@ class Grilo(GObject.GObject):
 
     @log
     def populate_artists(self, offset, callback, count=-1):
-        self.populate_items(Query.ARTISTS, offset, callback, count)
+        self.populate_items(Query.all_artists(), offset, callback, count)
 
     @log
     def populate_albums(self, offset, callback, count=-1):
-        self.populate_items(Query.ALBUMS, offset, callback, count)
+        self.populate_items(Query.all_albums(), offset, callback, count)
 
     @log
     def populate_songs(self, offset, callback, count=-1):
-        self.populate_items(Query.SONGS, offset, callback, count)
+        self.populate_items(Query.all_songs(), offset, callback, count)
 
     @log
-    def populate_album_songs(self, album_id, callback, count=-1):
-        self.populate_items(Query.album_songs(album_id), 0, callback, count)
+    def populate_album_songs(self, album, callback, count=-1):
+        if album.get_source() == 'grl-tracker-source':
+            self.populate_items(Query.album_songs(album.get_id()), 0, callback, count)
+        else:
+            source = self.sources[album.get_source()]
+            length = len(album.tracks)
+            for i, track in enumerate(album.tracks):
+                callback(source, None, track, length - (i + 1), None)
+            callback(source, None, None, 0, None)
 
     @log
-    def populate_items(self, query, offset, callback, count=-1):
+    def populate_custom_query(self, query, callback, count=-1, data=None):
+        self.populate_items(query, 0, callback, count, data)
+
+    @log
+    def populate_items(self, query, offset, callback, count=-1, data=None):
         options = self.options.copy()
         options.set_skip(offset)
         if count != -1:
             options.set_count(count)
 
-        def _callback(source, param, item, remaining, data, offset):
-            callback(source, param, item, remaining)
-        self.tracker.query(query, self.METADATA_KEYS, options, _callback, None)
+        def _callback(source, param, item, remaining, data, error):
+            callback(source, param, item, remaining, data)
+        self.tracker.query(query, self.METADATA_KEYS, options, _callback, data)
 
     @log
-    def _search_callback(self):
-        pass
-
-    @log
-    def search(self, q):
+    def search(self, q, callback, data=None):
         options = self.options.copy()
-        for source in self.sources:
-            logger.debug(source.get_name() + ' - ' + q)
-            source.search(q, [Grl.METADATA_KEY_ID], 0, 10,
-                          options, self._search_callback, source)
+
+        @log
+        def _search_callback(source, param, item, remaining, data, error):
+            callback(source, param, item, remaining, data)
+
+        if self.search_source:
+            if self.search_source.get_id().startswith('grl-upnp'):
+                options.set_type_filter(Grl.TypeFilter.AUDIO)
+            self.search_source.search(q, self.METADATA_KEYS, options,
+                                      _search_callback, data)
+        else:
+            Grl.multiple_search([self.sources[key] for key in self.sources
+                                 if key != 'grl-tracker-source'],
+                                q, self.METADATA_KEYS, options,
+                                _search_callback, data)
 
     @log
     def get_media_from_uri(self, uri, callback):
@@ -197,11 +236,19 @@ class Grilo(GObject.GObject):
         self.tracker.query(query, self.METADATA_KEYS, options, _callback, None)
 
     @log
-    def get_album_art_for_album_id(self, album_id, callback, data=None):
+    def get_album_art_for_item(self, item, callback, data=None):
+        item_id = item.get_id()
+
+        query = None
+        if isinstance(item, Grl.MediaAudio):
+            query = Query.get_album_for_song_id(item_id)
+        else:
+            query = Query.get_album_for_album_id(item_id)
+
         options = self.full_options.copy()
         options.set_count(1)
-        query = Query.get_album_for_id(album_id)
-        self.tracker.query(query, self.METADATA_THUMBNAIL_KEYS, options, callback, data)
+
+        self.search_source.query(query, self.METADATA_THUMBNAIL_KEYS, options, callback, data)
 
 
 Grl.init(None)
