@@ -113,7 +113,11 @@ class Player(GObject.GObject):
     def discover_item(self, item, callback, data=None):
         url = item.get_url()
         if not url:
-            print("The item %s doesn't have a URL set" % item)
+            logger.warn("The item %s doesn't have a URL set" % item)
+            return
+
+        if not url.startswith("file://"):
+            logger.log("Skipping discovery of %s as a remote url" % url)
             return
 
         obj = (callback, data)
@@ -159,13 +163,13 @@ class Player(GObject.GObject):
             uri = media.get_url()
         else:
             uri = 'none'
-            print('URI: ' + uri)
+        logger.warn('URI: ' + uri)
         error, debug = message.parse_error()
         debug = debug.split('\n')
         debug = [('     ') + line.lstrip() for line in debug]
         debug = '\n'.join(debug)
-        print('Error from element ' + message.src.get_name() + ': ' + error.message)
-        print('Debugging info:\n' + debug)
+        logger.warn('Error from element ' + message.src.get_name() + ': ' + error.message)
+        logger.warn('Debugging info:\n' + debug)
         self.play_next()
         return True
 
@@ -339,10 +343,18 @@ class Player(GObject.GObject):
 
     @log
     def _sync_playing(self):
-        image = self._pauseImage if self._get_playing() else self._playImage
+        if self._get_playing():
+            image = self._pauseImage
+            tooltip = _("Pause")
+        else:
+            image = self._playImage
+            tooltip = _("Play")
+
         if self.playBtn.get_image() != image:
             self.playBtn.set_image(image)
 
+        self.playBtn.set_tooltip_text(tooltip)
+
     @log
     def _sync_prev_next(self):
         hasNext = self.has_next()
@@ -380,6 +392,8 @@ class Player(GObject.GObject):
         try:
             assert media.get_artist() is not None
             artist = media.get_artist()
+        except:
+            pass
         finally:
             self.artistLabel.set_label(artist)
 
