@@ -43,6 +43,7 @@ from gi.repository import Tracker
 
 from gettext import gettext as _, ngettext
 from gnomemusic.grilo import grilo
+from gnomemusic.query import Query
 from gnomemusic.toolbar import ToolbarState
 import gnomemusic.widgets as Widgets
 from gnomemusic.playlists import Playlists
@@ -68,10 +69,9 @@ class ViewContainer(Gtk.Stack):
         nowPlayingIconName = 'media-playback-start-rtl-symbolic'
     errorIconName = 'dialog-error-symbolic'
     starIconName = 'starred-symbolic'
-    filter = None
 
     @log
-    def __init__(self, title, header_bar, selection_toolbar, view_type, use_sidebar=False, sidebar=None):
+    def __init__(self, name, title, window, view_type, use_sidebar=False, sidebar=None):
         Gtk.Stack.__init__(self,
                            transition_type=Gtk.StackTransitionType.CROSSFADE)
         self._grid = Gtk.Grid(orientation=Gtk.Orientation.HORIZONTAL)
@@ -99,10 +99,8 @@ class ViewContainer(Gtk.Stack):
             shadow_type=Gtk.ShadowType.NONE
         )
         self.view.set_view_type(view_type)
-        self.filter = self._model.filter_new(None)
-        self.view.set_model(self.filter)
+        self.view.set_model(self._model)
         self.vadjustment = self.view.get_vadjustment()
-        self.selection_toolbar = selection_toolbar
         box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
         box.pack_start(self.view, True, True, 0)
         if use_sidebar:
@@ -123,12 +121,15 @@ class ViewContainer(Gtk.Stack):
         self.view.connect('item-activated', self._on_item_activated)
         self.view.connect('selection-mode-request', self._on_selection_mode_request)
         self._cursor = None
-        self.header_bar = header_bar
+        self.window = window
+        self.header_bar = window.toolbar
+        self.selection_toolbar = window.selection_toolbar
         self.header_bar._select_button.connect(
             'toggled', self._on_header_bar_toggled)
         self.header_bar._cancel_button.connect(
             'clicked', self._on_cancel_button_clicked)
 
+        self.name = name
         self.title = title
         self.add(self._grid)
 
@@ -216,8 +217,9 @@ class ViewContainer(Gtk.Stack):
 
     @log
     def _connect_view(self):
-        self.view.connect_after('draw', self._on_view_draw)
-        self.view.add_events(Gdk.EventMask.EXPOSURE_MASK)
+        if self.view.get_view_type() == Gd.MainViewType.ICON:
+            self.view.connect_after('draw', self._on_view_draw)
+            self.view.add_events(Gdk.EventMask.EXPOSURE_MASK)
 
     @log
     def _on_view_draw(self, widget, cr):
@@ -275,7 +277,7 @@ class ViewContainer(Gtk.Stack):
             self._model.set(_iter, [8, 10], [self.errorIconName, True])
 
     @log
-    def _add_item(self, source, param, item, remaining):
+    def _add_item(self, source, param, item, remaining=0, data=None):
         if not item:
             return
         self._offset += 1
@@ -327,7 +329,7 @@ class ViewContainer(Gtk.Stack):
 # Class for the Empty View
 class Empty(Gtk.Stack):
     @log
-    def __init__(self, header_bar, player):
+    def __init__(self, window, player):
         Gtk.Stack.__init__(self,
                            transition_type=Gtk.StackTransitionType.CROSSFADE)
         builder = Gtk.Builder()
@@ -342,9 +344,8 @@ class Empty(Gtk.Stack):
 
 class Albums(ViewContainer):
     @log
-    def __init__(self, header_bar, selection_toolbar, player):
-        ViewContainer.__init__(self, _("Albums"), header_bar,
-                               selection_toolbar, Gd.MainViewType.ICON)
+    def __init__(self, window, player):
+        ViewContainer.__init__(self, 'albums', _("Albums"), window, Gd.MainViewType.ICON)
         self._albumWidget = Widgets.AlbumWidget(player)
         self.player = player
         self.add(self._albumWidget)
@@ -354,10 +355,16 @@ class Albums(ViewContainer):
 
     @log
     def _on_changes_pending(self, data=None):
-        if (self._init):
+        if (self._init and self.header_bar._selectionMode is False):
             self._offset = 0
             self._model.clear()
             self.populate()
+            grilo.changes_pending['Albums'] = False
+
+    @log
+    def _on_selection_mode_changed(self, widget, data=None):
+        if self.header_bar._selectionMode is False and grilo.changes_pending['Albums'] is True:
+            self._on_changes_pending()
 
     @log
     def _back_button_clicked(self, widget, data=None):
@@ -365,14 +372,13 @@ class Albums(ViewContainer):
 
     @log
     def _on_item_activated(self, widget, id, path):
-        child_path = self.filter.convert_path_to_child_path(path)
-        _iter = self._model.get_iter(child_path)
+        _iter = self._model.get_iter(path)
         title = self._model.get_value(_iter, 2)
         artist = self._model.get_value(_iter, 3)
         item = self._model.get_value(_iter, 5)
         self._albumWidget.update(artist, title, item,
                                  self.header_bar, self.selection_toolbar)
-        self.header_bar.set_state(0)
+        self.header_bar.set_state(ToolbarState.CHILD_VIEW)
         escaped_title = albumArtCache.get_media_title(item)
         self.header_bar.header_bar.set_title(escaped_title)
         self.header_bar.header_bar.sub_title = artist
@@ -385,7 +391,7 @@ class Albums(ViewContainer):
 
     @log
     def get_selected_track_uris(self, callback):
-        if self.header_bar._state == ToolbarState.SINGLE:
+        if self.header_bar._state == ToolbarState.CHILD_VIEW:
             uris = []
             for path in self._albumWidget.view.get_selection():
                 _iter = self._albumWidget.model.get_iter(path)
@@ -395,7 +401,7 @@ class Albums(ViewContainer):
             self.items_selected = []
             self.items_selected_callback = callback
             self.albums_index = 0
-            self.albums_selected = [self.filter.get_value(self.filter.get_iter(path), 5)
+            self.albums_selected = [self._model.get_value(self._model.get_iter(path), 5)
                                     for path in self.view.get_selection()]
             if len(self.albums_selected):
                 self._get_selected_album_songs()
@@ -403,12 +409,12 @@ class Albums(ViewContainer):
     @log
     def _get_selected_album_songs(self):
         grilo.populate_album_songs(
-            self.albums_selected[self.albums_index].get_id(),
+            self.albums_selected[self.albums_index],
             self._add_selected_item)
         self.albums_index += 1
 
     @log
-    def _add_selected_item(self, source, param, item, remaining):
+    def _add_selected_item(self, source, param, item, remaining=0, data=None):
         if item:
             self.items_selected.append(item.get_url())
         if remaining == 0:
@@ -420,8 +426,8 @@ class Albums(ViewContainer):
 
 class Songs(ViewContainer):
     @log
-    def __init__(self, header_bar, selection_toolbar, player):
-        ViewContainer.__init__(self, _("Songs"), header_bar, selection_toolbar, Gd.MainViewType.LIST)
+    def __init__(self, window, player):
+        ViewContainer.__init__(self, 'songs', _("Songs"), window, Gd.MainViewType.LIST)
         self._items = {}
         self.isStarred = None
         self.iter_to_clean = None
@@ -438,36 +444,39 @@ class Songs(ViewContainer):
 
     @log
     def _on_changes_pending(self, data=None):
-        if (self._init):
+        if (self._init and self.header_bar._selectionMode is False):
             self._model.clear()
             self._offset = 0
             self.populate()
+            grilo.changes_pending['Songs'] = False
+
+    @log
+    def _on_selection_mode_changed(self, widget, data=None):
+        if self.header_bar._selectionMode is False and grilo.changes_pending['Songs'] is True:
+            self._on_changes_pending()
 
     @log
     def _on_item_activated(self, widget, id, path):
-        _iter = self.filter.get_iter(path)
-        child_iter = self.filter.convert_iter_to_child_iter(_iter)
-        if self._model.get_value(child_iter, 8) != self.errorIconName:
-            self.player.set_playlist('Songs', None, self.filter, _iter, 5)
+        _iter = self._model.get_iter(path)
+        if self._model.get_value(_iter, 8) != self.errorIconName:
+            self.player.set_playlist('Songs', None, self._model, _iter, 5)
             self.player.set_playing(True)
 
     @log
     def update_model(self, player, playlist, currentIter):
         if self.iter_to_clean:
             self._model.set_value(self.iter_to_clean, 10, False)
-        if playlist != self.filter:
+        if playlist != self._model:
             return False
 
-        child_iter = self.filter.convert_iter_to_child_iter(currentIter)
-        self._model.set_value(child_iter, 10, True)
-        path = self._model.get_path(child_iter)
+        self._model.set_value(currentIter, 10, True)
+        path = self._model.get_path(currentIter)
         self.view.get_generic_view().scroll_to_path(path)
-        if self._model.get_value(child_iter, 8) != self.errorIconName:
-            self.iter_to_clean = child_iter.copy()
+        if self._model.get_value(currentIter, 8) != self.errorIconName:
+            self.iter_to_clean = currentIter.copy()
         return False
 
-    @log
-    def _add_item(self, source, param, item, remaining):
+    def _add_item(self, source, param, item, remaining=0, data=None):
         if not item:
             return
         self._offset += 1
@@ -475,6 +484,8 @@ class Songs(ViewContainer):
         artist = item.get_string(Grl.METADATA_KEY_ARTIST)\
             or item.get_author()\
             or _("Unknown Artist")
+        if item.get_url() is None:
+            return
         _iter = self._model.insert_with_valuesv(
             -1,
             [2, 3, 5, 8, 9, 10],
@@ -572,15 +583,15 @@ class Songs(ViewContainer):
 
     @log
     def get_selected_track_uris(self, callback):
-        callback([self.filter.get_value(self.filter.get_iter(path), 5).get_url()
+        callback([self._model.get_value(self._model.get_iter(path), 5).get_url()
                   for path in self.view.get_selection()])
 
 
 class Artists (ViewContainer):
     @log
-    def __init__(self, header_bar, selection_toolbar, player):
-        ViewContainer.__init__(self, _("Artists"), header_bar,
-                               selection_toolbar, Gd.MainViewType.LIST, True)
+    def __init__(self, window, player):
+        ViewContainer.__init__(self, 'artists', _("Artists"),
+                               window, Gd.MainViewType.LIST, True)
         self.artists_counter = 0
         self.player = player
         self._artists = {}
@@ -613,11 +624,12 @@ class Artists (ViewContainer):
 
     @log
     def _on_changes_pending(self, data=None):
-        if (self._init):
+        if (self._init and self.header_bar._selectionMode is False):
             self._model.clear()
             self._artists.clear()
             self._offset = 0
             self._populate()
+            grilo.changes_pending['Artists'] = False
 
     @log
     def _populate(self, data=None):
@@ -625,7 +637,7 @@ class Artists (ViewContainer):
         if not selection.get_selected()[1]:
             self._allIter = self._model.insert_with_valuesv(-1, [2], [_("All Artists")])
             self._last_selection = self._allIter
-            self._artists[_("All Artists").lower()] =\
+            self._artists[_("All Artists").casefold()] =\
                 {'iter': self._allIter, 'albums': [], 'widget': None}
             selection.select_path(self._model.get_path(self._allIter))
             self.view.emit('item-activated', '0',
@@ -654,13 +666,12 @@ class Artists (ViewContainer):
 
     @log
     def _on_item_activated(self, widget, item_id, path):
-        child_path = self.filter.convert_path_to_child_path(path)
-        _iter = self._model.get_iter(child_path)
+        _iter = self._model.get_iter(path)
         self._last_selection = _iter
         artist = self._model.get_value(_iter, 2)
-        albums = self._artists[artist.lower()]['albums']
+        albums = self._artists[artist.casefold()]['albums']
 
-        widget = self._artists[artist.lower()]['widget']
+        widget = self._artists[artist.casefold()]['widget']
         if widget:
             if widget.model == self.player.running_playlist('Artist', widget.artist):
                 self._artistAlbumsWidget = widget.get_parent()
@@ -682,10 +693,15 @@ class Artists (ViewContainer):
         artistAlbums = None
         if (self._model.get_string_from_iter(_iter) ==
                 self._model.get_string_from_iter(self._allIter)):
-            artistAlbums = Widgets.AllArtistsAlbums(self.player)
+            artistAlbums = Widgets.AllArtistsAlbums(
+                self.player, self.header_bar, self.selection_toolbar
+            )
         else:
-            artistAlbums = Widgets.ArtistAlbums(artist, albums, self.player)
-        self._artists[artist.lower()]['widget'] = artistAlbums
+            artistAlbums = Widgets.ArtistAlbums(
+                artist, albums, self.player,
+                self.header_bar, self.selection_toolbar
+            )
+        self._artists[artist.casefold()]['widget'] = artistAlbums
         new_artistAlbumsWidget.add(artistAlbums)
         new_artistAlbumsWidget.show()
 
@@ -694,18 +710,18 @@ class Artists (ViewContainer):
         GLib.idle_add(self.artistAlbumsStack.set_visible_child, new_artistAlbumsWidget)
 
     @log
-    def _add_item(self, source, param, item, remaining):
+    def _add_item(self, source, param, item, remaining=0, data=None):
         if item is None:
             return
         self._offset += 1
         artist = item.get_string(Grl.METADATA_KEY_ARTIST)\
             or item.get_author()\
             or _("Unknown Artist")
-        if not artist.lower() in self._artists:
+        if not artist.casefold() in self._artists:
             _iter = self._model.insert_with_valuesv(-1, [2], [artist])
-            self._artists[artist.lower()] = {'iter': _iter, 'albums': [], 'widget': None}
+            self._artists[artist.casefold()] = {'iter': _iter, 'albums': [], 'widget': None}
 
-        self._artists[artist.lower()]['albums'].append(item)
+        self._artists[artist.casefold()]['albums'].append(item)
 
     @log
     def populate(self):
@@ -730,6 +746,12 @@ class Artists (ViewContainer):
                 self.view.get_generic_view().get_selection().select_iter(
                     self._last_selection)
 
+    @log
+    def _on_selection_mode_changed(self, widget, data=None):
+        self.artistAlbumsStack.set_sensitive(not self.header_bar._selectionMode)
+        if self.header_bar._selectionMode is False and grilo.changes_pending['Artists'] is True:
+            self._on_changes_pending()
+
     @log
     def get_selected_track_uris(self, callback):
         self.items_selected = []
@@ -738,11 +760,11 @@ class Artists (ViewContainer):
         self.albums_selected = []
 
         for path in self.view.get_selection():
-            _iter = self.filter.get_iter(path)
-            artist = self.filter.get_value(_iter, 2)
-            albums = self._artists[artist.lower()]['albums']
-            if (self.filter.get_string_from_iter(_iter) !=
-                    self.filter.get_string_from_iter(self._allIter)):
+            _iter = self._model.get_iter(path)
+            artist = self._model.get_value(_iter, 2)
+            albums = self._artists[artist.casefold()]['albums']
+            if (self._model.get_string_from_iter(_iter) !=
+                    self._model.get_string_from_iter(self._allIter)):
                 self.albums_selected.extend(albums)
 
         if len(self.albums_selected):
@@ -751,12 +773,12 @@ class Artists (ViewContainer):
     @log
     def _get_selected_album_songs(self):
         grilo.populate_album_songs(
-            self.albums_selected[self.albums_index].get_id(),
+            self.albums_selected[self.albums_index],
             self._add_selected_item)
         self.albums_index += 1
 
     @log
-    def _add_selected_item(self, source, param, item, remaining):
+    def _add_selected_item(self, source, param, item, remaining=0, data=None):
         if item:
             self.items_selected.append(item.get_url())
         if remaining == 0:
@@ -770,13 +792,13 @@ class Playlist(ViewContainer):
     playlists_list = playlists.get_playlists()
 
     @log
-    def __init__(self, header_bar, selection_toolbar, player):
+    def __init__(self, window, player):
         self.playlists_sidebar = Gd.MainView(
             shadow_type=Gtk.ShadowType.NONE
         )
 
-        ViewContainer.__init__(self, _("Playlists"), header_bar,
-                               selection_toolbar, Gd.MainViewType.LIST, True, self.playlists_sidebar)
+        ViewContainer.__init__(self, 'playlists', _("Playlists"), window,
+                               Gd.MainViewType.LIST, True, self.playlists_sidebar)
 
         self.view.get_generic_view().get_style_context()\
             .add_class('songs-list')
@@ -950,13 +972,12 @@ class Playlist(ViewContainer):
     def update_model(self, player, playlist, currentIter):
         if self.iter_to_clean:
             self.iter_to_clean_model.set_value(self.iter_to_clean, 10, False)
-        if playlist != self.filter:
+        if playlist != self._model:
             return False
 
-        child_iter = self.filter.convert_iter_to_child_iter(currentIter)
-        self._model.set_value(child_iter, 10, True)
-        if self._model.get_value(child_iter, 8) != self.errorIconName:
-            self.iter_to_clean = child_iter.copy()
+        self._model.set_value(currentIter, 10, True)
+        if self._model.get_value(currentIter, 8) != self.errorIconName:
+            self.iter_to_clean = currentIter.copy()
             self.iter_to_clean_model = self._model
         return False
 
@@ -967,10 +988,9 @@ class Playlist(ViewContainer):
 
     @log
     def _on_item_activated(self, widget, id, path):
-        _iter = self.filter.get_iter(path)
-        child_iter = self.filter.convert_iter_to_child_iter(_iter)
-        if self._model.get_value(child_iter, 8) != self.errorIconName:
-            self.player.set_playlist('Playlist', self.current_playlist, self.filter, _iter, 5)
+        _iter = self._model.get_iter(path)
+        if self._model.get_value(_iter, 8) != self.errorIconName:
+            self.player.set_playlist('Playlist', self.current_playlist, self._model, _iter, 5)
             self.player.set_playing(True)
 
     @log
@@ -988,12 +1008,11 @@ class Playlist(ViewContainer):
         # use it as model, otherwise build the liststore
         cached_playlist = self.player.running_playlist('Playlist', playlist)
         if cached_playlist:
-            self._model = cached_playlist.get_model()
-            self.filter = cached_playlist
+            self._model = cached_playlist
             currentTrack = self.player.playlist.get_iter(self.player.currentTrack.get_path())
             self.update_model(self.player, cached_playlist,
                               currentTrack)
-            self.view.set_model(self.filter)
+            self.view.set_model(self._model)
             self.songs_count = self._model.iter_n_children(None)
             self._update_songs_count()
         else:
@@ -1010,14 +1029,13 @@ class Playlist(ViewContainer):
                 GObject.TYPE_BOOLEAN,
                 GObject.TYPE_BOOLEAN
             )
-            self.filter = self._model.filter_new(None)
-            self.view.set_model(self.filter)
+            self.view.set_model(self._model)
             playlists.parse_playlist(playlist, self._add_item)
             self.songs_count = 0
             self._update_songs_count()
 
     @log
-    def _add_item(self, source, param, item):
+    def _add_item(self, source, param, item, remaining=0, data=None):
         self._add_item_to_model(item, self._model)
 
     @log
@@ -1100,10 +1118,10 @@ class Playlist(ViewContainer):
 
     @log
     def _on_song_removed_from_playlist(self, playlists, name, uri):
+        cached_playlist = self.player.running_playlist('Playlist', name)
         if name == self.current_playlist:
             model = self._model
         else:
-            cached_playlist = self.player.running_playlist('Playlist', name)
             if cached_playlist and cached_playlist != self._model:
                 model = cached_playlist
             else:
@@ -1113,29 +1131,27 @@ class Playlist(ViewContainer):
         for row in model:
             if row[5].get_url() == uri:
                 # Is the removed track now being played?
-                if name == self.current_playlist:
+                if name == self.current_playlist and cached_playlist == model:
                     if self.player.currentTrack is not None:
                         currentTrackpath = self.player.currentTrack.get_path().to_string()
                         if row.path is not None and row.path.to_string() == currentTrackpath:
                             update_playing_track = True
 
+                nextIter = model.iter_next(row.iter)
                 model.remove(row.iter)
 
                 # Reload the model and switch to next song
                 if update_playing_track:
-                    if row.iter is None:
+                    if nextIter is None:
                         # Get first track if next track is not valid
-                        row.iter = model.get_iter_first()
-                        if row.iter is None:
+                        nextIter = model.get_iter_first()
+                        if nextIter is None:
                             # Last track was removed
                             return
 
                     self.iter_to_clean = None
-                    # row.iter will give us next iter to start playing
-                    # convert it to filter iter
-                    row.iter = self.filter.convert_child_iter_to_iter(row.iter)[1]
-                    self.update_model(self.player, self.filter, row.iter)
-                    self.player.set_playlist('Playlist', name, self.filter, row.iter, 5)
+                    self.update_model(self.player, model, nextIter)
+                    self.player.set_playlist('Playlist', name, model, nextIter, 5)
                     self.player.set_playing(True)
 
                 # Update songs count
@@ -1156,5 +1172,362 @@ class Playlist(ViewContainer):
 
     @log
     def get_selected_track_uris(self, callback):
-        callback([self.filter.get_value(self.filter.get_iter(path), 5).get_url()
+        callback([self._model.get_value(self._model.get_iter(path), 5).get_url()
                   for path in self.view.get_selection()])
+
+
+class Search(ViewContainer):
+    @log
+    def __init__(self, window, player):
+        ViewContainer.__init__(self, 'search', None, window, Gd.MainViewType.LIST)
+        self._items = {}
+        self.isStarred = None
+        self.iter_to_clean = None
+        self._iconHeight = 48
+        self._iconWidth = 48
+        self.cache = albumArtCache.get_default()
+        self._symbolicIcon = self.cache.get_default_icon(self._iconHeight,
+                                                         self._iconWidth)
+        self._add_list_renderers()
+        self.player = player
+        self.head_iters = [None, None, None, None]
+        self.songs_model = self._model
+
+        self.albums_selected = []
+        self._albums = {}
+        self._albumWidget = Widgets.AlbumWidget(player)
+        self.add(self._albumWidget)
+
+        self.artists_albums_selected = []
+        self._artists = {}
+        self._artistAlbumsWidget = None
+
+        self.view.get_generic_view().set_show_expanders(False)
+        self.items_selected = []
+        self.items_selected_callback = None
+
+    @log
+    def _back_button_clicked(self, widget, data=None):
+        if self.get_visible_child() == self._artistAlbumsWidget:
+            self._artistAlbumsWidget.destroy()
+            self._artistAlbumsWidget = None
+        self.window._stack.set_visible_child(self.window.prev_view)
+        self.set_visible_child(self._grid)
+
+    @log
+    def _on_item_activated(self, widget, id, path):
+        child_path = self.filter_model.convert_path_to_child_path(path)
+        _iter = self._model.get_iter(child_path)
+        if self._model[_iter][11] == 'album':
+            title = self._model.get_value(_iter, 2)
+            artist = self._model.get_value(_iter, 3)
+            item = self._model.get_value(_iter, 5)
+            self._albumWidget.update(artist, title, item,
+                                     self.header_bar, self.selection_toolbar)
+            self.header_bar.set_state(ToolbarState.SEARCH_VIEW)
+            escaped_title = albumArtCache.get_media_title(item)
+            self.header_bar.header_bar.set_title(escaped_title)
+            self.header_bar.header_bar.sub_title = artist
+            self.set_visible_child(self._albumWidget)
+            self.header_bar.searchbar.show_bar(False)
+        elif self._model[_iter][11] == 'artist':
+            artist = self._model.get_value(_iter, 2)
+            albums = self._artists[artist.casefold()]['albums']
+
+            self._artistAlbumsWidget = Widgets.ArtistAlbums(
+                artist, albums, self.player,
+                self.header_bar, self.selection_toolbar, True
+            )
+            self.add(self._artistAlbumsWidget)
+
+            self.header_bar.set_state(ToolbarState.SEARCH_VIEW)
+            self.header_bar.header_bar.set_title(artist)
+            self.set_visible_child(self._artistAlbumsWidget)
+            self.header_bar.searchbar.show_bar(False)
+        elif self._model[_iter][11] == 'song':
+            if self._model.get_value(_iter, 8) != self.errorIconName:
+                child_iter = self.songs_model.convert_child_iter_to_iter(_iter)[1]
+                self.player.set_playlist('Search Results', None, self.songs_model, child_iter, 5)
+                self.player.set_playing(True)
+        else:  # Headers
+            if self.view.get_generic_view().row_expanded(path):
+                self.view.get_generic_view().collapse_row(path)
+            else:
+                self.view.get_generic_view().expand_row(path, False)
+
+    @log
+    def _on_selection_mode_changed(self, widget, data=None):
+        if self._artistAlbumsWidget is not None and self.get_visible_child() == self._artistAlbumsWidget:
+            self._artistAlbumsWidget.set_selection_mode(self.header_bar._selectionMode)
+
+    @log
+    def _add_search_item(self, source, param, item, remaining=0, data=None):
+        if not item or data != self._model:
+            return
+
+        artist = item.get_string(Grl.METADATA_KEY_ARTIST) \
+            or item.get_author() \
+            or _("Unknown Artist")
+        album = item.get_string(Grl.METADATA_KEY_ALBUM) \
+            or _("Unknown Album")
+
+        key = '%s-%s' % (artist, album)
+        if key not in self._albums:
+            self._albums[key] = Grl.MediaBox()
+            self._albums[key].set_title(album)
+            self._albums[key].add_author(artist)
+            self._albums[key].set_source(source.get_id())
+            self._albums[key].tracks = []
+            self._add_item(source, None, self._albums[key], 0, [self._model, 'album'])
+            self._add_item(source, None, self._albums[key], 0, [self._model, 'artist'])
+
+        self._albums[key].tracks.append(item)
+        self._add_item(source, None, item, 0, [self._model, 'song'])
+
+    @log
+    def _add_item(self, source, param, item, remaining=0, data=None):
+        if data is None:
+            return
+
+        model, category = data
+        if not item or model != self._model:
+            return
+
+        self._offset += 1
+        title = albumArtCache.get_media_title(item)
+        item.set_title(title)
+        artist = item.get_string(Grl.METADATA_KEY_ARTIST) \
+            or item.get_author() \
+            or _("Unknown Artist")
+
+        group = 3
+        try:
+            group = {'album': 0, 'artist': 1, 'song': 2}[category]
+        except:
+            pass
+
+        _iter = None
+        if category == 'album' or category == 'song':
+            _iter = self._model.insert_with_values(
+                self.head_iters[group], -1,
+                [0, 2, 3, 4, 5, 8, 9, 10, 11],
+                [str(item.get_id()), title, artist,
+                 self._symbolicIcon, item, self.nowPlayingIconName,
+                 False, False, category])
+        else:
+            if not artist.casefold() in self._artists:
+                _iter = self._model.insert_with_values(
+                    self.head_iters[group], -1,
+                    [0, 2, 4, 5, 8, 9, 10, 11],
+                    [str(item.get_id()), artist,
+                     self._symbolicIcon, item, self.nowPlayingIconName,
+                     False, False, category])
+                self._artists[artist.casefold()] = {'iter': _iter, 'albums': []}
+
+            self._artists[artist.casefold()]['albums'].append(item)
+
+        if _iter:
+            albumArtCache.get_default().lookup(
+                item, self._iconWidth, self._iconHeight, self._on_lookup_ready,
+                _iter, artist, title)
+
+        if category == 'song':
+            self.player.discover_item(item, self._on_discovered, _iter)
+
+        if self._model.iter_n_children(self.head_iters[group]) == 1:
+            path = self._model.get_path(self.head_iters[group])
+            self.view.get_generic_view().expand_row(path, False)
+
+    @log
+    def _add_list_renderers(self):
+        list_widget = self.view.get_generic_view()
+        cols = list_widget.get_columns()
+
+        title_renderer = Gtk.CellRendererText(
+            xpad=12,
+            xalign=0.0,
+            yalign=0.5,
+            height=32,
+            ellipsize=Pango.EllipsizeMode.END,
+            weight=Pango.Weight.BOLD
+        )
+        list_widget.add_renderer(title_renderer,
+                                 self._on_list_widget_title_render, None)
+        cols[0].add_attribute(title_renderer, 'text', 2)
+
+        cells = cols[0].get_cells()
+        cols[0].reorder(cells[0], -1)
+        cols[0].set_cell_data_func(cells[0], self._on_list_widget_selection_render, None)
+
+    def _on_list_widget_selection_render(self, col, cell, model, _iter, data):
+        cell.set_visible(self.view.get_selection_mode() and model.iter_parent(_iter) is not None)
+
+    def _on_list_widget_title_render(self, col, cell, model, _iter, data):
+        cells = col.get_cells()
+        cells[0].set_visible(model.iter_parent(_iter) is not None)
+        cells[1].set_visible(model.iter_parent(_iter) is not None)
+        cells[2].set_visible(model.iter_parent(_iter) is None)
+
+    @log
+    def populate(self):
+        pass
+
+    @log
+    def get_selected_track_uris(self, callback):
+        if self.get_visible_child() == self._albumWidget:
+            uris = []
+            for path in self._albumWidget.view.get_selection():
+                _iter = self._albumWidget.model.get_iter(path)
+                uris.append(self._albumWidget.model.get_value(_iter, 5).get_url())
+            callback(uris)
+        elif self.get_visible_child() == self._artistAlbumsWidget:
+            uris = []
+            for row in self._artistAlbumsWidget.model:
+                if row[6]:
+                    uris.append(row[5].get_url())
+            callback(uris)
+        else:
+            self.items_selected = []
+            self.items_selected_callback = callback
+            self._get_selected_albums()
+
+    @log
+    def _get_selected_albums(self):
+        self.albums_index = 0
+        self.albums_selected = [self._model[child_path][5]
+                                for child_path in [self.filter_model.convert_path_to_child_path(path)
+                                                   for path in self.view.get_selection()]
+                                if self._model[child_path][11] == 'album']
+        if len(self.albums_selected):
+            self._get_selected_albums_songs()
+        else:
+            self._get_selected_artists()
+
+    @log
+    def _get_selected_albums_songs(self):
+        grilo.populate_album_songs(
+            self.albums_selected[self.albums_index],
+            self._add_selected_albums_songs)
+        self.albums_index += 1
+
+    @log
+    def _add_selected_albums_songs(self, source, param, item, remaining=0, data=None):
+        if item:
+            self.items_selected.append(item.get_url())
+        if remaining == 0:
+            if self.albums_index < len(self.albums_selected):
+                self._get_selected_albums_songs()
+            else:
+                self._get_selected_artists()
+
+    @log
+    def _get_selected_artists(self):
+        self.artists_albums_index = 0
+        self.artists_selected = [self._artists[self._model[child_path][2].casefold()]
+                                 for child_path in [self.filter_model.convert_path_to_child_path(path)
+                                                    for path in self.view.get_selection()]
+                                 if self._model[child_path][11] == 'artist']
+
+        self.artists_albums_selected = []
+        for artist in self.artists_selected:
+            self.artists_albums_selected.extend(artist['albums'])
+
+        if len(self.artists_albums_selected):
+            self._get_selected_artists_albums_songs()
+        else:
+            self._get_selected_songs()
+
+    @log
+    def _get_selected_artists_albums_songs(self):
+        grilo.populate_album_songs(
+            self.artists_albums_selected[self.artists_albums_index],
+            self._add_selected_artists_albums_songs)
+        self.artists_albums_index += 1
+
+    @log
+    def _add_selected_artists_albums_songs(self, source, param, item, remaining=0, data=None):
+        if item:
+            self.items_selected.append(item.get_url())
+        if remaining == 0:
+            if self.artists_albums_index < len(self.artists_albums_selected):
+                self._get_selected_artists_albums_songs()
+            else:
+                self._get_selected_songs()
+
+    @log
+    def _get_selected_songs(self):
+        self.items_selected.extend([self._model[child_path][5].get_url()
+                                    for child_path in [self.filter_model.convert_path_to_child_path(path)
+                                                       for path in self.view.get_selection()]
+                                    if self._model[child_path][11] == 'song'])
+        self.items_selected_callback(self.items_selected)
+
+    @log
+    def _filter_visible_func(self, model, _iter, data=None):
+        return model.iter_parent(_iter) is not None or model.iter_has_child(_iter)
+
+    @log
+    def set_search_text(self, search_term, fields_filter):
+        query_matcher = {
+            'album': {
+                'search_all': Query.get_albums_with_any_match,
+                'search_artist': Query.get_albums_with_artist_match,
+                'search_album': Query.get_albums_with_album_match,
+                'search_track': Query.get_albums_with_track_match,
+            },
+            'artist': {
+                'search_all': Query.get_artists_with_any_match,
+                'search_artist': Query.get_artists_with_artist_match,
+                'search_album': Query.get_artists_with_album_match,
+                'search_track': Query.get_artists_with_track_match,
+            },
+            'song': {
+                'search_all': Query.get_songs_with_any_match,
+                'search_artist': Query.get_songs_with_artist_match,
+                'search_album': Query.get_songs_with_album_match,
+                'search_track': Query.get_songs_with_track_match,
+            },
+        }
+
+        self._model = Gtk.TreeStore(
+            GObject.TYPE_STRING,
+            GObject.TYPE_STRING,
+            GObject.TYPE_STRING,    # item title or header text
+            GObject.TYPE_STRING,    # artist for albums and songs
+            GdkPixbuf.Pixbuf,       # album art
+            GObject.TYPE_OBJECT,    # item
+            GObject.TYPE_BOOLEAN,
+            GObject.TYPE_INT,
+            GObject.TYPE_STRING,
+            GObject.TYPE_BOOLEAN,
+            GObject.TYPE_BOOLEAN,
+            GObject.TYPE_STRING     # type
+        )
+        self.filter_model = self._model.filter_new(None)
+        self.filter_model.set_visible_func(self._filter_visible_func)
+        self.view.set_model(self.filter_model)
+
+        self._albums = {}
+        self._artists = {}
+
+        if search_term == "":
+            return
+
+        albums_iter = self._model.insert_with_values(None, -1, [2], [_("Albums")])
+        artists_iter = self._model.insert_with_values(None, -1, [2], [_("Artists")])
+        songs_iter = self._model.insert_with_values(None, -1, [2], [_("Songs")])
+        playlists_iter = self._model.insert_with_values(None, -1, [2], [_("Playlists")])
+
+        self.head_iters = [albums_iter, artists_iter, songs_iter, playlists_iter]
+        self.songs_model = self._model.filter_new(self._model.get_path(songs_iter))
+
+        # Use queries for Tracker
+        if not grilo.search_source or \
+           grilo.search_source.get_id() == 'grl-tracker-source':
+            for category in ('album', 'artist', 'song'):
+                query = query_matcher[category][fields_filter](search_term)
+                grilo.populate_custom_query(query, self._add_item, -1, [self._model, category])
+        if not grilo.search_source or \
+           grilo.search_source.get_id() != 'grl-tracker-source':
+            # nope, can't do - reverting to Search
+            grilo.search(search_term, self._add_search_item, self._model)
