@@ -31,7 +31,7 @@
 # delete this exception statement from your version.
 
 
-from gi.repository import Gtk, Gd, GLib, GObject, Pango
+from gi.repository import Gtk, Gdk, Gd, GLib, GObject, Pango
 from gi.repository import GdkPixbuf, Gio, Grl
 from gi.repository import Tracker
 from gettext import gettext as _, ngettext
@@ -64,7 +64,6 @@ class AlbumWidget(Gtk.EventBox):
     tracks = []
     duration = 0
     symbolicIcon = ALBUM_ART_CACHE.get_default_icon(256, 256)
-    filter = None
 
     @log
     def __init__(self, player):
@@ -81,11 +80,13 @@ class AlbumWidget(Gtk.EventBox):
         )
         self.view.set_view_type(Gd.MainViewType.LIST)
         self.album = None
+        self.header_bar = None
         self.view.connect('item-activated', self._on_item_activated)
         self.monitors = []
         view_box = self.ui.get_object('view')
         self.ui.get_object('scrolledWindow').set_placement(Gtk.CornerType.
                                                            TOP_LEFT)
+        self.view.connect('selection-mode-request', self._on_selection_mode_request)
         child_view = self.view.get_children()[0]
         child_view.set_margin_top(64)
         child_view.set_margin_bottom(64)
@@ -99,10 +100,14 @@ class AlbumWidget(Gtk.EventBox):
         self.get_style_context().add_class('content-view')
         self.show_all()
 
+    @log
+    def _on_selection_mode_request(self, *args):
+        self.header_bar._select_button.clicked()
+
     @log
     def _on_item_activated(self, widget, id, path):
-        child_path = self.filter.convert_path_to_child_path(path)
-        _iter = self.model.get_iter(child_path)
+        _iter = self.model.get_iter(path)
+
         if(self.model.get_value(_iter, 7) != ERROR_ICON_NAME):
             if (self.iterToClean and self.player.playlistId == self.album):
                 item = self.model.get_value(self.iterToClean, 5)
@@ -193,16 +198,14 @@ class AlbumWidget(Gtk.EventBox):
         else:
             self.duration = 0
             self._create_model()
-            GLib.idle_add(grilo.populate_album_songs, item.get_id(),
-                          self._on_populate_album_songs)
+            GLib.idle_add(grilo.populate_album_songs, item, self.add_item)
         header_bar._select_button.connect(
             'toggled', self._on_header_select_button_toggled)
         header_bar._cancel_button.connect(
             'clicked', self._on_header_cancel_button_clicked)
         self.view.connect('view-selection-changed',
                           self._on_view_selection_changed)
-        self.filter = self.model.filter_new(None)
-        self.view.set_model(self.filter)
+        self.view.set_model(self.model)
         escaped_artist = GLib.markup_escape_text(artist)
         escaped_album = GLib.markup_escape_text(album)
         self.ui.get_object('artist_label').set_markup(escaped_artist)
@@ -254,7 +257,7 @@ class AlbumWidget(Gtk.EventBox):
             self.model.set(_iter, [7, 9], [ERROR_ICON_NAME, True])
 
     @log
-    def _on_populate_album_songs(self, source, prefs, track, remaining):
+    def add_item(self, source, prefs, track, remaining, data=None):
         if track:
             self.tracks.append(track)
             self.duration = self.duration + track.get_duration()
@@ -324,11 +327,16 @@ class AlbumWidget(Gtk.EventBox):
 class ArtistAlbums(Gtk.VBox):
 
     @log
-    def __init__(self, artist, albums, player):
+    def __init__(self, artist, albums, player,
+                 header_bar, selection_toolbar, selectionModeAllowed=False):
         Gtk.VBox.__init__(self)
         self.player = player
         self.artist = artist
         self.albums = albums
+        self.selectionMode = False
+        self.selectionModeAllowed = selectionModeAllowed
+        self.selection_toolbar = selection_toolbar
+        self.header_bar = header_bar
         self.ui = Gtk.Builder()
         self.ui.add_from_resource('/org/gnome/Music/ArtistAlbumsWidget.ui')
         self.set_border_width(0)
@@ -343,6 +351,7 @@ class ArtistAlbums(Gtk.VBox):
                                    GObject.TYPE_OBJECT,   # song object
                                    GObject.TYPE_BOOLEAN
                                    )
+        self.model.connect('row-changed', self._model_row_changed)
 
         self._hbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
         self._albumBox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
@@ -365,7 +374,10 @@ class ArtistAlbums(Gtk.VBox):
 
     @log
     def add_album(self, album):
-        widget = ArtistAlbumWidget(self.artist, album, self.player, self.model)
+        widget = ArtistAlbumWidget(
+            self.artist, album, self.player, self.model,
+            self.header_bar, self.selectionModeAllowed
+        )
         self._albumBox.pack_start(widget, False, False, 0)
         self.widgets.append(widget)
 
@@ -419,12 +431,35 @@ class ArtistAlbums(Gtk.VBox):
             itr = self.model.iter_next(itr)
         return False
 
+    @log
+    def set_selection_mode(self, selectionMode):
+        self.selectionMode = selectionMode
+        for widget in self.widgets:
+            widget.set_selection_mode(selectionMode)
+
+    @log
+    def _model_row_changed(self, model, path, _iter):
+        if not self.selectionMode:
+            return
+        selected_items = 0
+        for row in model:
+            if row[6]:
+                selected_items += 1
+        self.selection_toolbar\
+            ._add_to_playlist_button.set_sensitive(selected_items > 0)
+        if selected_items > 0:
+            self.header_bar._selection_menu_label.set_text(
+                ngettext("Selected %d item", "Selected %d items", selected_items) % selected_items)
+        else:
+            self.header_bar._selection_menu_label.set_text(_("Click on items to select them"))
+
 
 class AllArtistsAlbums(ArtistAlbums):
 
     @log
-    def __init__(self, player):
-        ArtistAlbums.__init__(self, _("All Artists"), [], player)
+    def __init__(self, player, header_bar, selection_toolbar, selectionModeAllowed=False):
+        ArtistAlbums.__init__(self, _("All Artists"), [], player,
+                              header_bar, selection_toolbar, selectionModeAllowed)
         self._offset = 0
         self._populate()
 
@@ -435,7 +470,7 @@ class AllArtistsAlbums(ArtistAlbums):
                           self._offset, self.add_item, -1)
 
     @log
-    def add_item(self, source, param, item, remaining):
+    def add_item(self, source, param, item, remaining=0, data=None):
         if item:
             self._offset += 1
             self.add_album(item)
@@ -444,12 +479,16 @@ class AllArtistsAlbums(ArtistAlbums):
 class ArtistAlbumWidget(Gtk.HBox):
 
     @log
-    def __init__(self, artist, album, player, model):
+    def __init__(self, artist, album, player, model, header_bar, selectionModeAllowed):
         super(Gtk.HBox, self).__init__()
         self.player = player
         self.album = album
         self.artist = artist
         self.model = model
+        self.model.connect('row-changed', self._model_row_changed)
+        self.header_bar = header_bar
+        self.selectionMode = False
+        self.selectionModeAllowed = selectionModeAllowed
         self.songs = []
         self.monitors = []
         self.ui = Gtk.Builder()
@@ -467,8 +506,7 @@ class ArtistAlbumWidget(Gtk.HBox):
                 str(album.get_creation_date().get_year())
             )
         self.tracks = []
-        GLib.idle_add(grilo.populate_album_songs,
-                      album.get_id(), self.get_songs)
+        GLib.idle_add(grilo.populate_album_songs, album, self.add_item)
         self.pack_start(self.ui.get_object('ArtistAlbumWidget'), True, True, 0)
         self.show_all()
 
@@ -483,7 +521,7 @@ class ArtistAlbumWidget(Gtk.HBox):
             song_widget.can_be_played = False
 
     @log
-    def get_songs(self, source, prefs, track, remaining):
+    def add_item(self, source, prefs, track, remaining, data=None):
         if track:
             self.tracks.append(track)
         else:
@@ -508,6 +546,11 @@ class ArtistAlbumWidget(Gtk.HBox):
                 song_widget._iter = itr
                 song_widget.model = self.model
                 song_widget.title = ui.get_object('title')
+                song_widget.checkButton = ui.get_object('select')
+                song_widget.checkButton.set_visible(self.selectionMode)
+                song_widget.checkButton.connect(
+                    'toggled', self._check_button_toggled, song_widget
+                )
                 self.player.discover_item(track, self._on_discovered, song_widget)
                 g_file = Gio.file_new_for_uri(track.get_url())
                 self.monitors.append(g_file.monitor_file(Gio.FileMonitorFlags.NONE,
@@ -549,15 +592,49 @@ class ArtistAlbumWidget(Gtk.HBox):
             self.ui.get_object('cover').set_from_pixbuf(pixbuf)
 
     @log
-    def track_selected(self, widget, _iter):
+    def track_selected(self, widget, event):
         if not widget.can_be_played:
             return
 
+        if not self.selectionMode and \
+            (event.button == Gdk.BUTTON_SECONDARY or
+                (event.button == 1 and event.state & Gdk.ModifierType.CONTROL_MASK)):
+            if self.selectionModeAllowed:
+                self.header_bar._select_button.set_active(True)
+            else:
+                return
+
+        if self.selectionMode:
+            self.model[widget._iter][6] = not self.model[widget._iter][6]
+            return
+
         self.player.stop()
         self.player.set_playlist('Artist', self.artist,
                                  widget.model, widget._iter, 5)
         self.player.set_playing(True)
 
+    @log
+    def set_selection_mode(self, selectionMode):
+        self.selectionMode = selectionMode
+        for songWidget in self.songs:
+            songWidget.checkButton.set_visible(selectionMode)
+            if not selectionMode:
+                songWidget.model[songWidget._iter][6] = False
+
+    @log
+    def _check_button_toggled(self, button, songWidget):
+        if songWidget.model[songWidget._iter][6] != button.get_active():
+            songWidget.model[songWidget._iter][6] = button.get_active()
+
+    @log
+    def _model_row_changed(self, model, path, _iter):
+        if not model[_iter][5]:
+            return
+        songWidget = model[_iter][5].song_widget
+        selected = model[_iter][6]
+        if selected != songWidget.checkButton.get_active():
+            songWidget.checkButton.set_active(selected)
+
 
 class PlaylistDialog():
     @log
