@@ -73,6 +73,9 @@ class Window(Gtk.ApplicationWindow):
         self.set_size_request(200, 100)
         self.set_icon_name('gnome-music')
 
+        self.prev_view = None
+        self.curr_view = None
+
         size_setting = self.settings.get_value('window-size')
         if isinstance(size_setting[0], int) and isinstance(size_setting[1], int):
             self.resize(size_setting[0], size_setting[1])
@@ -107,21 +110,21 @@ class Window(Gtk.ApplicationWindow):
     @log
     def _on_changes_pending(self, data=None):
         count = 1
-        cursor = tracker.query(Query.SONGS_COUNT, None)
+        cursor = tracker.query(Query.all_songs_count(), None)
         if cursor is not None and cursor.next(None):
             count = cursor.get_integer(0)
         if not count > 0:
-            print("switching to Empty view")
-            self._stack.disconnect(self._on_notify_model_id)
-            self.disconnect(self._key_press_event_id)
-            for i in range(0, 4):
-                view = self.views.pop()
-                view.destroy()
-            self.toolbar.hide_stack()
-            self._switch_to_empty_view()
+            if self.toolbar._selectionMode is False and len(self.views) != 1:
+                self._stack.disconnect(self._on_notify_model_id)
+                self.disconnect(self._key_press_event_id)
+                view_count = len(self.views)
+                for i in range(0, view_count):
+                    view = self.views.pop()
+                    view.destroy()
+                self.toolbar.hide_stack()
+                self._switch_to_empty_view()
         else:
             if (self.views[0] == self.views[-1]):
-                print("switching to player view")
                 view = self.views.pop()
                 view.destroy()
                 self._switch_to_player_view()
@@ -184,21 +187,24 @@ class Window(Gtk.ApplicationWindow):
             transition_duration=100,
             visible=True,
             can_focus=False)
+        self._overlay = Gtk.Overlay(child=self._stack)
+        self._overlay.add_overlay(self.toolbar.dropdown)
         self.set_titlebar(self.toolbar.header_bar)
         self._box.pack_start(self.toolbar.searchbar, False, False, 0)
-        self._box.pack_start(self._stack, True, True, 0)
+        self._box.pack_start(self._overlay, True, True, 0)
         self._box.pack_start(self.player.eventBox, False, False, 0)
         self._box.pack_start(self.selection_toolbar.eventbox, False, False, 0)
         self.add(self._box)
         count = 1
-        cursor = tracker.query(Query.SONGS_COUNT, None)
+        cursor = tracker.query(Query.all_songs_count(), None)
         if cursor is not None and cursor.next(None):
             count = cursor.get_integer(0)
         if count > 0:
             self._switch_to_player_view()
         # To revert to the No Music View when no songs are found
         else:
-            self._switch_to_empty_view()
+            if self.toolbar._selectionMode is False:
+                self._switch_to_empty_view()
 
         self.toolbar._search_button.connect('toggled', self._on_search_toggled)
         self.toolbar.connect('selection-mode-changed', self._on_selection_mode_changed)
@@ -207,52 +213,66 @@ class Window(Gtk.ApplicationWindow):
         self.selection_toolbar._remove_from_playlist_button.connect(
             'clicked', self._on_remove_from_playlist_button_clicked)
 
-        self.toolbar.set_state(ToolbarState.ALBUMS)
+        self.toolbar.set_state(ToolbarState.MAIN)
         self.toolbar.header_bar.show()
+        self._overlay.show()
         self.player.eventBox.show_all()
         self._box.show()
         self.show()
 
     @log
     def _switch_to_empty_view(self):
-        self.views.append(Views.Empty(self.toolbar, self.player))
+        self.views.append(Views.Empty(self, self.player))
         self._stack.add_titled(self.views[0], _("Empty"), _("Empty"))
         self.toolbar._search_button.set_sensitive(False)
         self.toolbar._select_button.set_sensitive(False)
 
     @log
     def _switch_to_player_view(self):
-        self.views.append(Views.Albums(self.toolbar, self.selection_toolbar, self.player))
-        self.views.append(Views.Artists(self.toolbar, self.selection_toolbar, self.player))
-        self.views.append(Views.Songs(self.toolbar, self.selection_toolbar, self.player))
-        self.views.append(Views.Playlist(self.toolbar, self.selection_toolbar, self.player))
-
-        for i in self.views:
-            self._stack.add_titled(i, i.title, i.title)
-
-        self.toolbar.set_stack(self._stack)
-        self.toolbar.searchbar.show()
-
         self._on_notify_model_id = self._stack.connect('notify::visible-child', self._on_notify_mode)
         self.connect('destroy', self._notify_mode_disconnect)
         self._key_press_event_id = self.connect('key_press_event', self._on_key_press)
 
+        self.views.append(Views.Albums(self, self.player))
+        self.views.append(Views.Artists(self, self.player))
+        self.views.append(Views.Songs(self, self.player))
+        self.views.append(Views.Playlist(self, self.player))
+        self.views.append(Views.Search(self, self.player))
+
+        for i in self.views:
+            if i.title:
+                self._stack.add_titled(i, i.name, i.title)
+            else:
+                self._stack.add_named(i, i.name)
+
+        self.toolbar.set_stack(self._stack)
+        self.toolbar.searchbar.show()
+        self.toolbar.dropdown.show()
+
         self.views[0].populate()
 
+    @log
+    def _set_selection(self, model, value, parent=None):
+        count = 0
+        _iter = model.iter_children(parent)
+        while _iter is not None:
+            if model.iter_has_child(_iter):
+                count += self._set_selection(model, value, _iter)
+            if model[_iter][5]:
+                model.set(_iter, [6], [value])
+                count += 1
+            _iter = model.iter_next(_iter)
+        return count
+
     @log
     def _on_select_all(self, action, param):
         if self.toolbar._selectionMode is False:
             return
-        if self.toolbar._state != ToolbarState.SINGLE:
+        if self.toolbar._state == ToolbarState.MAIN:
             model = self._stack.get_visible_child()._model
         else:
-            model = self._stack.get_visible_child()._albumWidget.model
-        _iter = model.get_iter_first()
-        count = 0
-        while _iter is not None:
-            model.set(_iter, [6], [True])
-            _iter = model.iter_next(_iter)
-            count = count + 1
+            model = self._stack.get_visible_child().get_visible_child().model
+        count = self._set_selection(model, True)
         if count > 0:
             self.toolbar._selection_menu_label.set_text(
                 ngettext("Selected %d item", "Selected %d items", count) % count)
@@ -264,16 +284,13 @@ class Window(Gtk.ApplicationWindow):
 
     @log
     def _on_select_none(self, action, param):
-        if self.toolbar._state != ToolbarState.SINGLE:
+        if self.toolbar._state == ToolbarState.MAIN:
             model = self._stack.get_visible_child()._model
         else:
-            model = self._stack.get_visible_child()._albumWidget.model
-        _iter = model.get_iter_first()
+            model = self._stack.get_visible_child().get_visible_child().model
+        self._set_selection(model, False)
         self.selection_toolbar._add_to_playlist_button.set_sensitive(False)
         self.selection_toolbar._remove_from_playlist_button.set_sensitive(False)
-        while _iter is not None:
-            model.set(_iter, [6], [False])
-            _iter = model.iter_next(_iter)
         self.toolbar._selection_menu_label.set_text(_("Click on items to select them"))
         self._stack.get_visible_child().queue_draw()
 
@@ -314,12 +331,16 @@ class Window(Gtk.ApplicationWindow):
 
     @log
     def _on_notify_mode(self, stack, param):
-        # Slide out artist list on switching to artists view
-        if stack.get_visible_child() == self.views[1] or \
-           stack.get_visible_child() == self.views[3]:
-            stack.get_visible_child().stack.set_visible_child_name('dummy')
-            stack.get_visible_child().stack.set_visible_child_name('sidebar')
-        self.toolbar.searchbar.show_bar(False)
+        self.prev_view = self.curr_view
+        self.curr_view = stack.get_visible_child()
+
+        # Slide out sidebar on switching to Artists or Playlists view
+        if self.curr_view == self.views[1] or \
+           self.curr_view == self.views[3]:
+            self.curr_view.stack.set_visible_child_name('dummy')
+            self.curr_view.stack.set_visible_child_name('sidebar')
+        if self.curr_view != self.views[4]:
+            self.toolbar.searchbar.show_bar(False)
 
     @log
     def _toggle_view(self, btn, i):
@@ -328,10 +349,17 @@ class Window(Gtk.ApplicationWindow):
     @log
     def _on_search_toggled(self, button, data=None):
         self.toolbar.searchbar.show_bar(button.get_active())
+        if not button.get_active() and self.curr_view == self.views[4] and \
+           self.toolbar._state == ToolbarState.MAIN:
+            self._stack.set_visible_child(self.prev_view)
+            if self.toolbar._selectionMode:
+                self.toolbar.set_selection_mode(False)
 
     @log
     def _on_selection_mode_changed(self, widget, data=None):
-        if self.toolbar._selectionMode:
+        if self.toolbar._selectionMode is False:
+            self._on_changes_pending()
+        else:
             in_playlist = self._stack.get_visible_child() == self.views[3]
             self.selection_toolbar._add_to_playlist_button.set_visible(not in_playlist)
             self.selection_toolbar._remove_from_playlist_button.set_visible(in_playlist)
